// ----------------------------------------------------------------------------
// M o d u l e . c p p   f o r   i i r . s e p
// ----------------------------------------------------------------------------
// second order iir filter module for synthedit
// author: Rudolf Lindner, rlindner@gmx.net
// april, 2003
// ----------------------------------------------------------------------------
// this is freeware, unauthorized commercial use is not permitted
// ----------------------------------------------------------------------------


#include "Module.h"
#include "SEMod_struct.h"
#include "SEPin.h"
#include <assert.h>

#include <stdio.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <winbase.h>


// handy function to fix denormals. don't work on P4's
static const float anti_denormal = 1e-18f;
inline void kill_denormal(float &p_sample ) { p_sample += anti_denormal;p_sample -= anti_denormal;};

#define PN_OUTPUT  6
#define PN_INPUT   0

static char Id [] = "IIR-Filter Synthedit Plugin - rlindner@gmx.net";

Module::Module(seaudioMasterCallback seaudioMaster) : SEModule_base(seaudioMaster)
{
}

Module::~Module()
{
}




void Module::sub_process(long buffer_offset, long sampleFrames )
{
    float *input    = buffer_offset + input1;
    float *output   = buffer_offset + output1;

    float *pa1      = buffer_offset + koeff_a1;
    float *pa2      = buffer_offset + koeff_a2;
    float *pb0      = buffer_offset + koeff_b0;
    float *pb1      = buffer_offset + koeff_b1;
    float *pb2      = buffer_offset + koeff_b2;
    

	for( int s = 0; s < sampleFrames ; s++ )
	{
        
        // scale values SE maps range 0..10 (volts) to 0.0 .. 1.0
        float a1  = *pa1++ * 10;
        float a2  = *pa2++ * 10;
        float b0  = *pb0++ * 10;
        float b1  = *pb1++ * 10;
        float b2  = *pb2++ * 10;

        float in  = *input++;
        kill_denormal(in);



        /* Filter network:

                               w(n)
        in -->(+)------------+-----(*)----(*)--> out
               ^             |      |      ^
               |             v      b0     |
               |           [1/z]           |
               |             |             |
               |             | w(n-1)      |
              (+)<---(*)-----+-----(*)----(*)
               |      |      |      |      ^
               |      a1     v      b1     |
               |           [1/z]           |
               |             |             |
               |             | w(n-2)      |
               +-----(*)-----+-----(*)-----+
                      |             |
                      a2            b2
        
        (+):  adder
        (*):  multiplier
        1/z:  one sample delay
        w(k): intermediate samples
        a(k): coefficients feedback path
        b(k): coefficients forward path
        
        */



        float out = b0 * wn + b1 * wn_1 + b2 * wn_2;
        wn        = in      + a1 * wn_1 + a2 * wn_2;


        wn_2      = wn_1;
        wn_1      = wn;

        *output++ = out;        
	}
}

// perform any initialisation here
void Module::open()
{
    SET_PROCESS_FUNC(Module::sub_process);
	SEModule_base::open(); 
    
    wn = wn_1 = wn_2 = (float)0;
	
    // let 'downstream' modules know audio data is coming
	getPin(PN_OUTPUT)->TransmitStatusChange( SampleClock(), ST_RUN );
}


void Module::OnPlugStateChange(SEPin *pin)
{

}


// describe the module
bool Module::getModuleProperties (SEModuleProperties* properties)
{
	// describe the plugin, this is the name the end-user will see.
	properties->name = "2nd order IIR/FIR Filter network";
    
	// return a unique string 32 characters max
	// if posible include manufacturer and plugin identity
	// this is used internally by SE to identify the plug. No two plugs may have the same id.
	properties->id = "r|ma labs iir";

	// Info, may include Author, Web page whatever
	properties->about = "\
2nd order iir/fir filter network______________________________________\n\n\
    aj: coefficients feedback path\n\
    bj: coefficients forward path\n\n\
\n\
    Transfer function H(z) = (b0 + b1*z^-1 + b2*z^-2) / (a0 + a1*z^-1)\n\n\
______________________________________created by rlindner@gmx.net";

	return true;
}


// describe the pins (plugs)
bool Module::getPinProperties (long index, SEPinProperties* properties)
{
	switch( index )
	{
	case 0:
		properties->name				= "input";
		properties->variable_address	= &input1;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case 1:
		properties->name				= "a1";
		properties->variable_address	= &koeff_a1;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case 2:
		properties->name				= "a2";
		properties->variable_address	= &koeff_a2;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case 3:
		properties->name				= "b0";
		properties->variable_address	= &koeff_b0;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case 4:
		properties->name				= "b1";
		properties->variable_address	= &koeff_b1;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case 5:
		properties->name				= "b2";
		properties->variable_address	= &koeff_b2;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case 6:
		properties->name				= "output";
		properties->variable_address	= &output1;
		properties->direction			= DR_OUT;
		properties->datatype			= DT_FSAMPLE;
		break;
    default:
		return false; // host will ask for plugs 0,1,2,3 etc. return false to signal when done
	};

	return true;
}


